
#include "docopt/docopt.h"

#include <ace/Init_ACE.h>
#include "Future_orig.h"
#include "Future_fixed.h"
#include <ace/Time_Value.h>
#include <ace/Countdown_Time.h>

#include <string>
#include <iostream>
#include <sstream>
#include <vector>
#include <thread>
#include <random>
#include <condition_variable>
#include <mutex>

static const char USAGE[] =
R"(X-Gen worker process.

usage:
  acefuturetester [-f] [-d DURATION] [-t NUMTHREADS] [-v]
  acefuturetester (-h | --help)

options:
  -h --help           Print help
  -f                  test the changed implementation without it will be the original implementation
  -d DURATION         how long the test should run (default: 14400 i.e. 4 hours)
  -t NUMTHREADS       number of test threads to start (default: 1)
  -v                  verbose, enable debugging
)";


void acefuture_runner( time_t Duration, bool Verbose )
{
  std::random_device              rd;
  std::mt19937                    gen(rd());
  std::uniform_int_distribution<> dis(1,1000000);
  ACE_Time_Value duration( Duration );
  ACE_Countdown_Time timer(&duration);
  timer.start();
  uint64_t runNum = 0;
  do
  {
    if( ++runNum % 5000 == 0 )
    {
      std::cout << "futurerunner: thread = " << std::this_thread::get_id() << ", time to go = " << duration << ": iteration #" << runNum << "\n";
    }
    int a = dis(gen);
    int b = dis(gen);
    int c = dis(gen);
    int expected_res = a+b*c;
    int actual_res;
    ACE_Future_Orig<int> result;
    std::thread cal { [](int u, int v, int w, ACE_Future_Orig<int> Result) { int r=u+v*w; Result.set(r); }, a, b, c, result };
    result.get(actual_res);
    if( actual_res != expected_res ) abort(); // hit the bug...
    cal.join();
    timer.update();
  } while( duration != ACE_Time_Value::zero );
  std::cout << "futurerunner: thread = " << std::this_thread::get_id() << ", finished after = " << Duration << ": with " << runNum << " iterations done\n";
}

void acefuture_fixed_runner( time_t Duration, bool Verbose )
{
  std::random_device              rd;
  std::mt19937                    gen(rd());
  std::uniform_int_distribution<> dis(1,1000000);
  ACE_Time_Value duration( Duration );
  ACE_Countdown_Time timer(&duration);
  timer.start();
  uint64_t runNum = 0;
  do
  {
    if( ++runNum % 5000 == 0 )
    {
      std::cout << "future_fixed_runner: thread = " << std::this_thread::get_id() << ", time to go = " << duration << ": iteration #" << runNum << "\n";
    }
    int a = dis(gen);
    int b = dis(gen);
    int c = dis(gen);
    int expected_res = a+b*c;
    int actual_res;
    ACE_Future_fixed<int> result;
    std::thread cal { [](int u, int v, int w, ACE_Future_fixed<int> Result) { int r=u+v*w; Result.set(r); }, a, b, c, result };
    result.get(actual_res);
    if( actual_res != expected_res ) abort(); // hit the bug...
    cal.join();
    timer.update();
  } while( duration != ACE_Time_Value::zero );
  std::cout << "future_fixed_runner: thread = " << std::this_thread::get_id() << ", finished after = " << Duration << ": with " << runNum << " iterations done\n";
}


int main(int argc, const char** argv )
{
  docopt::Options args;
  try
  {
    args= docopt::docopt_parse(USAGE, { argv + 1, argv + argc }, /*handle help*/ true, "acefuturetester 1.0");
  }
  catch (const docopt::DocoptExitHelp&)
  {
    std::cout << USAGE;
    return 0;
  }
  catch (const docopt::DocoptExitVersion&)
  {
    std::cout << "acefuturetester 1.0";
    return 0;
  }
  catch (const std::exception& ex)
  {
    std::cout << ex.what() << std::endl;
    return 1;
  }

  ACE::init();
  long duration = 3600*4; // 4 hours
  long numThreads = 2;
  bool verbose = false;
  bool testFix = false;
  try
  {
    for(auto const& arg : args)
    {
      //std::cout << arg.first << " -> " <<  arg.second << std::endl;
           if (arg.first == "-d" && arg.second.isString()) duration   = arg.second.asLong();
      else if (arg.first == "-t" && arg.second.isString()) numThreads = arg.second.asLong();
      else if (arg.first == "-v" && arg.second.isBool()  ) verbose    = arg.second.asBool();
      else if (arg.first == "-f" && arg.second.isBool())   testFix    = arg.second.asBool();
    }
  }
  catch (const std::exception& ex)
  {
    std::cout << ex.what() << std::endl;
    ACE::fini();
    return 1;
  }

  std::vector<std::thread> threads;
  threads.reserve(numThreads);
  for( long i = 0; i < numThreads; ++i )
  {
    if( testFix )
    {
      std::cout << "Starting thread " << i << ".\n";
      threads.emplace_back( std::thread { acefuture_fixed_runner, duration, verbose } );
    }
    else
    {
      std::cout << "Starting thread " << i << ".\n";
      threads.emplace_back( std::thread { acefuture_runner, duration, verbose } );
    }
  }
  std::cout << "All threads started waiting for them to finish.\n";
  for( auto& thr : threads )
  {
    thr.join();
  }
  std::cout << "All threads finished, cleanup and exit.\n";



  ACE::fini();
  return 0;
}
